/+
 ' Source          : InNumRange.d v1.0 - is the numerical value in range.
 ' Author          : David L. 'SpottedTiger' Davis
 ' D Compiler      : dmd v0.130
 ' Source Platform : Intel x86 / WinXP SP2
 ' Target Platform : An Intel x86-Based OS like MS Windows, also it  
 '                 : would work on others CPUs if the isNegative() 
 '                 : function were adjusted to work with that CPU.
 '                 :
 ' Created         : 08.Sep.05 Compiled and tested with dmd v0.130.
 ' Modified        : 10.Sep.05 Improved using range checking against
 '                 :           floating-point datatypes, and added the
 '                 :           isWholeNumber() template function.
 '                 : 19.Sep.05 Added a BigEndian version block
 '                 :           in the isNegative() function.
 '                 : 24.Oct.05 Compiled and tested with dmd v0.137.
 '                 : 30.Dec.05 Compiled and tested with dmd v0.142.
 '                 : 23.Mar.06 Compiled and tested with dmd v0.150.
 '                 : 28.May.06 Recompiled and Tested with dmd v0.159.
 '                 : 19.Jul.06 All "Imports" are now private by default  
 '                 :           (which shouldn't be a problem, since I  
 '                 :           always use the private keyword) in 
 '                 :           dmd v0.163, recompiled and tested.
 '                 : 26.Aug.06 Added comparing whole numbers against
 '                 :           a floating-point whole numbers (fraction 
 '                 :           is zero). And compiled and tested with
 '                 :           dmd v0.165
 '                 : 08.Sep.06 Added isNumValueInRange(...) as the main
 '                 :           function, which now calls the isValueInRange()
 '                 :           as a private function. Compiled and 
 '                 :           tested with dmd v0.166.
 '                 : 05.Nov.06 Compiled and tested with dmd v0.173.
 '                 : 25.Nov.06 Compiled and tested with dmd v0.175.
 '                 : 09.Dec.06 Compiled and tested with dmd v0.177.
 '                 : 05.Jan.07 Compiled and tested with dmd v1.0.
 '                 : 09.Feb.07 Compiled and tested with dmd v1.005. 
 '                 : 03.Jan.08 Compiled and tested with dmd v1.025.  
 '                 : 20.Mar.08 Compiled and tested with dmd v1.028. 
 '                 : 19.Jun.08 Compiled and tested with dmd v1.031.
 '                 : 17.Jul.08 Compiled and tested with dmd v1.033/v1.032
 '                 :           Entries like typeof(T2) v2 now gets an error 
 '                 :           and will need to be written as T2 v2 to work
 '                 : 13.May.09 Compiled and tested with dmd v1.045.
 '                 :
 ' Licence         : Public Domain / Contributed to Digital Mars
 ' -------------------------------------------------------------------
 ' To unittest: dmd innumrange.d -debug=inrange -unittest
 ' To compile : dmd YourSource.d innumrange.d
 +/
 module innumrange;

debug( inrange ) private import scomp = std.compiler; 
debug( inrange ) private import std.stdio;
private import std.math;
private import std.string; // for tolower()
private import std.stdarg; // for argument gathering
private import std.utf;    // for toUTF8()
debug( inrange ) private import std.conv;

// Setup TypeoF const types
const char[]  TF_CHARA;
const void[]  TF_VOIDA;
const wchar[] TF_WCHARA;
const dchar[] TF_DCHARA;
const char    TF_CHAR;
const wchar   TF_WCHAR;
const dchar   TF_DCHAR;
const bool    TF_BOOL;
const byte    TF_BYTE;
const ubyte   TF_UBYTE;
const short   TF_SHORT;
const ushort  TF_USHORT;
const int     TF_INT;
const uint    TF_UINT;
const long    TF_LONG;
const ulong   TF_ULONG;
const float   TF_FLOAT;
const double  TF_DOUBLE;
const real    TF_REAL;
const ifloat  TF_IFLOAT;
const idouble TF_IDOUBLE;
const ireal   TF_IREAL;
const cfloat  TF_CFLOAT;
const cdouble TF_CDOUBLE;
const creal   TF_CREAL;
//const cent    TF_CENT;
//const ucent   TF_UCENT;

/+
 ' isNegative!(T)(Tvalue)
 ' --------------------------------------
 ' T : the datatype of the input value
 ' [in] v : is the value being passed in,
 '          and compared to see if it is a
 '          negative number.
 ' 
 ' bool return : returns true if the value  
 '               is a negative number.
 '
 ' Note: That char, wchar, and dchar are unsigned
 '       according the the D "Basic Data Types"
 '       descriptions. But in "Integer Promotions"  
 '       char and wchar are implicitly converted 
 '       to an int, whereas dchar is implicitly 
 '       converted to a uint. 
 +/
template isNegative(T)
{
    bool isNegative(in T v)
    {
        TypeInfo ti = typeid(T);

        if ((ti is typeid(bool))   || 
            (ti is typeid(ubyte))  ||
            (ti is typeid(ushort)) ||
            (ti is typeid(uint))   ||
            (ti is typeid(ulong))  ||
            (ti is typeid(char))   ||
            (ti is typeid(wchar))  ||
            (ti is typeid(dchar)))
            return false;
        else if ((ti is typeid(byte))    ||
                 (ti is typeid(short))   ||
                 (ti is typeid(int))     ||
                 (ti is typeid(long))    ||
                 (ti is typeid(float))   ||
                 (ti is typeid(double))  ||
                 (ti is typeid(real))    ||
                 (ti is typeid(ifloat))  ||
                 (ti is typeid(idouble)) ||
                 (ti is typeid(ireal)))
        {
            // Code borrowed from the std.math.signbit() 
            ubyte* psign = cast(ubyte*)&v;

            // For non-Intel x86 CPUs that use BigEndian byte order
            version( BigEndian )
            {
                return (psign[0] & 0x80) != 0;
            }
            else // defaults to x86's LittleEndian byte order
            {
                uint ui = ti.tsize - 1;
                return (psign[ui] & 0x80) != 0;
            }
            //^-^-^-^-^-^-^-^-^-^-^-^-^-^-^-^-^-^-^
        }
        else
            return false;
    }
}

/+
 ' isWholeNumber!(T)(Tvalue)
 ' --------------------------------------
 ' T : the datatype of the input value
 ' [in] v : is the value being passed in,
 '          and compared to see if it is 
 '          datatype and/or value is a whole number.
 ' 
 ' bool return : returns true if the datatype   
 '               and/or value is a whole number,   
 '               float must have a zero remainer.
 '
+/
template isWholeNumber(T)
{
    bool isWholeNumber(in T v)
    {
        if (typeid(T) is typeid(bool)   ||
            typeid(T) is typeid(byte)   ||
            typeid(T) is typeid(ubyte)  || 
            typeid(T) is typeid(short)  ||
            typeid(T) is typeid(ushort) || 
            typeid(T) is typeid(int)    ||
            typeid(T) is typeid(uint)   || 
            typeid(T) is typeid(long)   ||
            typeid(T) is typeid(ulong)  ||
            typeid(T) is typeid(char)   ||
            typeid(T) is typeid(wchar)  ||
            typeid(T) is typeid(dchar))
        {
            return true;
        }          
        else if (typeid(T) is typeid(float)   ||
                 typeid(T) is typeid(double)  ||
                 typeid(T) is typeid(real))
        {
            real rRemainer = 0.0L;
            
            //std.stdio.writefln("isWholeNumber: v=%f, cast(real)v=%f", v, cast(ifloat)v);
            if (v == 0.0)
                return true;
                
            std.math.modf(cast(real)v, rRemainer);
            
            /+
            // testing code
            std.stdio.writefln("std.math.modf(%f, rRemainer))=%f", 
                                cast(real)v, rRemainer);
            +/
                                 
            if (rRemainer != 0.0L && rRemainer == v)
                return true;
            else
                return false;
        }
        else if (typeid(T) is typeid(ifloat)  ||
                 typeid(T) is typeid(idouble) ||
                 typeid(T) is typeid(ireal)) 
        {
            real rRemainer = 0.0L;
            real rValue    = cast(real)(v * -1.0i); 
            //std.stdio.writefln("isWholeNumber: typeid(T)=%s, v=%gi, rValue=%f", 
            //                   typeid(T), v, rValue);
            if (rValue == 0.0L)
                return true;
                
            std.math.modf(rValue, rRemainer);

            /+
            debug( inrange )
            {
            if (rRemainer != 0.0L && rRemainer == rValue)
                std.stdio.writefln("isWholeNumber: v=%gi, true", v);
            else
                std.stdio.writefln("isWholeNumber: v=%gi, false", v);
            }
            +/
            
            if (rRemainer != 0.0L && rRemainer == rValue)
                return true;
            else
                return false;
        }
        else if (typeid(T) is typeid(cfloat)  ||
                 typeid(T) is typeid(cdouble) ||
                 typeid(T) is typeid(creal))    
        { 
            return false;
        }
        else
            return false;
    }
}

/+
 ' isValueInRange!(T1, T2)(T1value)
 ' --------------------------------------
 ' T1 : the datatype of the input value
 ' T2 : the base datatype comparing range against
 ' T1value : is the value being passed in.
 ' 
 ' bool return : returns true if the value is 
 '               within the base datatype's range.
 '
 ' Notes :  Compare any Integer value against a 
 ' numeric datatype's min / max, except for the 
 ' floating-point datatypes.
 +/
template isValueInRange(T1, T2)
{
     private bool isValueInRange(in T1 v)
     {   
     	 /+
         debug( inrange )
         {
             writefln("T1.typid=%s, T1.v=", typeid(T1), v,
                      " T1.tsize=%d, isNeg=%b, isWhole=%b ", 
                       T1.sizeof, isNegative(v), isWholeNumber(v), 
                      "T2.typid=%s, T2.tsize=%d", typeid(T2), T2.sizeof); 
         }     
         +/
         
         //writefln(
         bool bTVneg = isNegative(v);
         bool bTBneg = isNegative(T2.min);
         
         //real rRemainer = 0.0L;
         
         // If same datatypes, they're already in range of each other.
         if (typeid(T1) is typeid(T2))
         {
             debug( inrange ) writefln("Values are the exact same datatypes"c);
             return true;
         }
         else if (typeid(T2) is typeid(bool)   ||
                  typeid(T2) is typeid(byte)   ||
                  typeid(T2) is typeid(ubyte)  || 
                  typeid(T2) is typeid(short)  ||
                  typeid(T2) is typeid(ushort) || 
                  typeid(T2) is typeid(int)    ||
                  typeid(T2) is typeid(uint)   || 
                  typeid(T2) is typeid(long)   ||
                  typeid(T2) is typeid(ulong)  ||
                  typeid(T2) is typeid(char)   ||
                  typeid(T2) is typeid(wchar)  ||
                  typeid(T2) is typeid(dchar))
         {
             if (typeid(T1) is typeid(float)   ||
                 typeid(T1) is typeid(double)  || 
                 typeid(T1) is typeid(real))
             {
                 debug( inrange ) writefln("Floating-point Value to an Integer Base"c);
                 if (isWholeNumber(v))
                 {
                     //typeof(T2) v2; // for dmd v1.031 and below
                     T2 v2; //for dmd v1.033/v1.032 new way 
                     
                     //if (v2 == v)
                     if (v >= v2.min && v <= v2.max)
                         return true;	
                     else
                         return false;    
                 }    
                 else
                     return false;
             }  
             else if (typeid(T1) is typeid(ifloat)  || 
                      typeid(T1) is typeid(idouble) ||
                      typeid(T1) is typeid(ireal))
             {
                 debug( inrange ) writefln("Imaginary Floating-point Value to an Integer Base"c);
                 if (isWholeNumber(v))
                 {
                     //typeof(T2) v2 = cast(T2)(v * -1.0i); //for dmd v1.031 and below
                     T2 v2 = cast(T2)(v * -1.0i); //for dmd v1.033/v1.032 new way
                     
                     //writefln("isValueInRange: typeid(T2)=%s", typeid(T2), ", 
                     //         "v2=", v2, ", v=", v, ", v2==v=", (v2 * 1.0i) == v);
                     if ((v2 * 1.0i) == v || v2 * -1.0i == v)
                         return true;	
                     else
                         return false;    
                 }    
                 else
                     return false;
             	
             }
             
             // fall thru to the logic below;  
         }
         else if (typeid(T2) is typeid(float))
         {
             debug( inrange ) writefln("Float Value to an Integer Base"c);
             float f = cast(float)v;
             if (f == v)
             //std.math.modf(cast(real)v, rRemainer);
             //if (rRemainer == 0.0L)
                 return true;
             else 
                 return false; 
         }
         else if (typeid(T2) is typeid(ifloat))
         {
             debug( inrange ) writefln("IFloat Value to an Integer Base"c);
             ifloat ift = cast(ifloat)(v * 1.0i);
             //writefln("isValueInRange: ift=%gi, v=", ift, v, ", ift==v=", ift == v * 1.0i);
             if (ift == v * 1.0i)
                 return true;
             else 
                 return false; 
         }
         else if (typeid(T2) is typeid(double))
         {
             debug( inrange ) writefln("Double Value to an Integer Base"c);
             double d = cast(double)v;
             if (d == v)
                 return true;
             else 
                 return false; 
         }
         else if (typeid(T2) is typeid(idouble))
         {
             debug( inrange ) writefln("IDouble Value to an Integer Base"c);
             idouble id = cast(idouble)(v * 1.0i);
             if (id == v * 1.0i)
                 return true;
             else 
                 return false; 
         }
         else if (typeid(T2) is typeid(real))
         {
             debug( inrange ) writefln("Real Value to an Integer Base"c);
             real r = cast(real)v;
             if (r == v)
                 return true;
             else
                 return false; 
         }
         else if (typeid(T2) is typeid(ireal))
         {
             debug( inrange ) writefln("IReal Value to an Integera Base"c);
             ireal ir = cast(ireal)(v * 1.0i);
             if (ir == v * 1.0i)
                 return true;
             else
                 return false; 
         }
         else if (typeid(T1) is typeid(cfloat)  ||
                  typeid(T1) is typeid(cdouble) ||
                  typeid(T1) is typeid(creal))
         {
             debug( inrange ) writefln("Complex Float Value to a Base"c);
             return false;
         }

         // Negative Value to Positive Base
         if (bTVneg && !bTBneg)
         {
             debug( inrange ) writefln("Negative Value to Positive Base"c);
             return false;
         }
         // Positive Value to Positive Base
         else if (!bTVneg && !bTBneg)
         {
             debug( inrange ) writefln("Positive Value to Positive Base"c);
             if (isWholeNumber(v))
             {
                 ulong v2 = cast(ulong)v;
                 ulong ul = cast(ulong)T2.max;

                 if (v2 >= 0 && v2 <= ul)
                     return true;
                 else
                     return false;
            }
            else
                return false;
         }
         // Positive Value to a Negative Base
         // (we can ignore below the zero range)
         else if (!bTVneg && bTBneg) 
         {
             debug( inrange ) writefln("Positive Value to a Negative Base"c);
             if (isWholeNumber(v))
             {
                 ulong v2 = cast(ulong)v;
                 ulong ul = cast(ulong)T2.max;

                 if (v2 >= 0 && v2 <= ul)
                     return true;
                 else
                     return false; 
             }
             else
                 return false;
         }
         // Negative Value to a Negative Base
         // (we can move both into the positive number range)
         else if (bTVneg && bTBneg) 
         {
             debug( inrange ) writefln("Negative Value to a Negative Base"c);
             if (isWholeNumber(v))
             {
                 long l  = cast(long)T2.min;
                 long v2 = cast(long)v;
                 v2 = l - v2;

                 if (isNegative(v2) || v2 == 0)
                     return true;
                 else
                     return false;
             }
             else
                 return false;
         }
         else
         {
            debug( inrange ) writefln("Other"c);
            return false;
         }
     }
}

// Allow any object as a parameter, but only test number values (excluding complex floating-points).
// char, wchar, and dchar are also numbers when they're passsed into this function.
bool isNumValueInRange(...)
{
    if (_arguments.length < 2)
        return false;
        
    TypeInfo ti1 = _arguments[0];
    TypeInfo ti2 = _arguments[1];

    /+
    debug( inrange )
    {
        writefln("ti1.typeid=%s, ti2.typeid=%s, equals=%b", ti1, ti2, ti1 is ti2); 
    }     
    +/
    
    bool isValidDatatype(TypeInfo ti)
    {
        if (ti !is typeid(bool)    &&
            ti !is typeid(byte)    &&
            ti !is typeid(ubyte)   && 
            ti !is typeid(short)   &&
            ti !is typeid(ushort)  && 
            ti !is typeid(int)     &&
            ti !is typeid(uint)    && 
            ti !is typeid(long)    &&
            ti !is typeid(ulong)   &&
            ti !is typeid(float)   &&
            ti !is typeid(double)  && 
            ti !is typeid(real)    &&
            ti !is typeid(ifloat)  &&
            ti !is typeid(idouble) && 
            ti !is typeid(ireal)   &&
            ti !is typeid(char)    &&
            ti !is typeid(wchar)   &&
            ti !is typeid(dchar)) 
        {
            //writefln("ti datatype not supported");     
            return false;    
        }
        else 
            return true;
    }
    
    if (!isValidDatatype(ti1))
        return false;
        
    if (!isValidDatatype(ti2))
        return false; 
           
    if (ti1 is ti2) // same datatypes
    {
        //writefln("ti1 is ti2; same datatypes");
        return true;
    }    

    if (ti1 is typeid(bool))
    {
        bool b = (va_arg!(bool)(_argptr));

        if (ti2 is typeid(bool))
            return isValueInRange!(bool, bool)(b);
        else if (ti2 is typeid(byte))
            return isValueInRange!(bool, byte)(b);
        else if (ti2 is typeid(ubyte))
            return isValueInRange!(bool, ubyte)(b); 
        else if(ti2 is typeid(short))
            return isValueInRange!(bool, short)(b);
        else if (ti2 is typeid(ushort))
            return isValueInRange!(bool, ushort)(b); 
        else if(ti2 is typeid(int))
            return isValueInRange!(bool, int)(b);
        else if (ti2 is typeid(uint))
            return isValueInRange!(bool, uint)(b); 
        else if (ti2 is typeid(long))
            return isValueInRange!(bool, long)(b);
        else if (ti2 is typeid(ulong))
            return isValueInRange!(bool, ulong)(b);
        else if (ti2 is typeid(float)) 
            return isValueInRange!(bool, float)(b);
        else if (ti2 is typeid(double))
            return isValueInRange!(bool, double)(b); 
        else if (ti2 is typeid(real))
            return isValueInRange!(bool, real)(b);
        else if (ti2 is typeid(ifloat))
            return isValueInRange!(bool, ifloat)(b);
        else if (ti2 is typeid(idouble))
            return isValueInRange!(bool, idouble)(b); 
        else if (ti2 is typeid(ireal))
            return isValueInRange!(bool, ireal)(b);    	
    }
    else if (ti1 is typeid(char))     
    {
        char c = (va_arg!(char)(_argptr)); 
        if (ti2 is typeid(bool))
            return isValueInRange!(char, bool)(c);
        else if (ti2 is typeid(byte))
            return isValueInRange!(char, byte)(c);
        else if (ti2 is typeid(ubyte))
            return isValueInRange!(char, ubyte)(c); 
        else if(ti2 is typeid(short))
            return isValueInRange!(char, short)(c);
        else if (ti2 is typeid(ushort))
            return isValueInRange!(char, ushort)(c); 
        else if(ti2 is typeid(int))
            return isValueInRange!(char, int)(c);
        else if (ti2 is typeid(uint))
            return isValueInRange!(char, uint)(c); 
        else if (ti2 is typeid(long))
            return isValueInRange!(char, long)(c);
        else if (ti2 is typeid(ulong))
            return isValueInRange!(char, ulong)(c);
        else if (ti2 is typeid(float)) 
            return isValueInRange!(char, float)(c);
        else if (ti2 is typeid(double))
            return isValueInRange!(char, double)(c); 
        else if (ti2 is typeid(real))
            return isValueInRange!(char, real)(c);
        else if (ti2 is typeid(ifloat))
            return isValueInRange!(char, ifloat)(c);
        else if (ti2 is typeid(idouble))
            return isValueInRange!(char, idouble)(c); 
        else if (ti2 is typeid(ireal))
            return isValueInRange!(char, ireal)(c);    	        
    }    
    else if (ti1 is typeid(wchar))     
    {
        wchar wc = (va_arg!(wchar)(_argptr)); 
        if (ti2 is typeid(bool))
            return isValueInRange!(wchar, bool)(wc);
        else if (ti2 is typeid(byte))
            return isValueInRange!(wchar, byte)(wc);
        else if (ti2 is typeid(ubyte))
            return isValueInRange!(wchar, ubyte)(wc); 
        else if(ti2 is typeid(short))
            return isValueInRange!(wchar, short)(wc);
        else if (ti2 is typeid(ushort))
            return isValueInRange!(wchar, ushort)(wc); 
        else if(ti2 is typeid(int))
            return isValueInRange!(wchar, int)(wc);
        else if (ti2 is typeid(uint))
            return isValueInRange!(wchar, uint)(wc); 
        else if (ti2 is typeid(long))
            return isValueInRange!(wchar, long)(wc);
        else if (ti2 is typeid(ulong))
            return isValueInRange!(wchar, ulong)(wc);
        else if (ti2 is typeid(float)) 
            return isValueInRange!(wchar, float)(wc);
        else if (ti2 is typeid(double))
            return isValueInRange!(wchar, double)(wc); 
        else if (ti2 is typeid(real))
            return isValueInRange!(wchar, real)(wc);
        else if (ti2 is typeid(ifloat))
            return isValueInRange!(wchar, ifloat)(wc);
        else if (ti2 is typeid(idouble))
            return isValueInRange!(wchar, idouble)(wc); 
        else if (ti2 is typeid(ireal))
            return isValueInRange!(wchar, ireal)(wc);    	        
    }        
    else if (ti1 is typeid(dchar))     
    {
        dchar dc = (va_arg!(dchar)(_argptr)); 
        if (ti2 is typeid(bool))
            return isValueInRange!(dchar, bool)(dc);
        else if (ti2 is typeid(byte))
            return isValueInRange!(dchar, byte)(dc);
        else if (ti2 is typeid(ubyte))
            return isValueInRange!(dchar, ubyte)(dc); 
        else if(ti2 is typeid(short))
            return isValueInRange!(dchar, short)(dc);
        else if (ti2 is typeid(ushort))
            return isValueInRange!(dchar, ushort)(dc); 
        else if(ti2 is typeid(int))
            return isValueInRange!(dchar, int)(dc);
        else if (ti2 is typeid(uint))
            return isValueInRange!(dchar, uint)(dc); 
        else if (ti2 is typeid(long))
            return isValueInRange!(dchar, long)(dc);
        else if (ti2 is typeid(ulong))
            return isValueInRange!(dchar, ulong)(dc);
        else if (ti2 is typeid(float)) 
            return isValueInRange!(dchar, float)(dc);
        else if (ti2 is typeid(double))
            return isValueInRange!(dchar, double)(dc); 
        else if (ti2 is typeid(real))
            return isValueInRange!(dchar, real)(dc);
        else if (ti2 is typeid(ifloat))
            return isValueInRange!(dchar, ifloat)(dc);
        else if (ti2 is typeid(idouble))
            return isValueInRange!(dchar, idouble)(dc); 
        else if (ti2 is typeid(ireal))
            return isValueInRange!(dchar, ireal)(dc);    	        
    }            
    else if (ti1 is typeid(byte))  
    {   
        byte byt = (va_arg!(byte)(_argptr)); 
        if (ti2 is typeid(bool))
            return isValueInRange!(byte, bool)(byt);
        else if (ti2 is typeid(byte))
            return isValueInRange!(byte, byte)(byt);
        else if (ti2 is typeid(ubyte))
            return isValueInRange!(byte, ubyte)(byt); 
        else if(ti2 is typeid(short))
            return isValueInRange!(byte, short)(byt);
        else if (ti2 is typeid(ushort))
            return isValueInRange!(byte, ushort)(byt); 
        else if(ti2 is typeid(int))
            return isValueInRange!(byte, int)(byt);
        else if (ti2 is typeid(uint))
            return isValueInRange!(byte, uint)(byt); 
        else if (ti2 is typeid(long))
            return isValueInRange!(byte, long)(byt);
        else if (ti2 is typeid(ulong))
            return isValueInRange!(byte, ulong)(byt);
        else if (ti2 is typeid(float)) 
            return isValueInRange!(byte, float)(byt);
        else if (ti2 is typeid(double))
            return isValueInRange!(byte, double)(byt); 
        else if (ti2 is typeid(real))
            return isValueInRange!(byte, real)(byt);
        else if (ti2 is typeid(ifloat))
            return isValueInRange!(byte, ifloat)(byt);
        else if (ti2 is typeid(idouble))
            return isValueInRange!(byte, idouble)(byt); 
        else if (ti2 is typeid(ireal))
            return isValueInRange!(byte, ireal)(byt);    	          
    }        
    else if (ti1 is typeid(ubyte))  
    {   
        ubyte ubyt = (va_arg!(ubyte)(_argptr)); 
        if (ti2 is typeid(bool))
            return isValueInRange!(ubyte, bool)(ubyt);
        else if (ti2 is typeid(byte))
            return isValueInRange!(ubyte, byte)(ubyt);
        else if (ti2 is typeid(ubyte))
            return isValueInRange!(ubyte, ubyte)(ubyt); 
        else if(ti2 is typeid(short))
            return isValueInRange!(ubyte, short)(ubyt);
        else if (ti2 is typeid(ushort))
            return isValueInRange!(ubyte, ushort)(ubyt); 
        else if(ti2 is typeid(int))
            return isValueInRange!(ubyte, int)(ubyt);
        else if (ti2 is typeid(uint))
            return isValueInRange!(ubyte, uint)(ubyt); 
        else if (ti2 is typeid(long))
            return isValueInRange!(ubyte, long)(ubyt);
        else if (ti2 is typeid(ulong))
            return isValueInRange!(ubyte, ulong)(ubyt);
        else if (ti2 is typeid(float)) 
            return isValueInRange!(ubyte, float)(ubyt);
        else if (ti2 is typeid(double))
            return isValueInRange!(ubyte, double)(ubyt); 
        else if (ti2 is typeid(real))
            return isValueInRange!(ubyte, real)(ubyt);
        else if (ti2 is typeid(ifloat))
            return isValueInRange!(ubyte, ifloat)(ubyt);
        else if (ti2 is typeid(idouble))
            return isValueInRange!(ubyte, idouble)(ubyt); 
        else if (ti2 is typeid(ireal))
            return isValueInRange!(ubyte, ireal)(ubyt);    	          
    }    
    else if (ti1 is typeid(short))     
    {
        short sht = (va_arg!(short)(_argptr)); 
        if (ti2 is typeid(bool))
            return isValueInRange!(short, bool)(sht);
        else if (ti2 is typeid(byte))
            return isValueInRange!(short, byte)(sht);
        else if (ti2 is typeid(ubyte))
            return isValueInRange!(short, ubyte)(sht); 
        else if(ti2 is typeid(short))
            return isValueInRange!(short, short)(sht);
        else if (ti2 is typeid(ushort))
            return isValueInRange!(short, ushort)(sht); 
        else if(ti2 is typeid(int))
            return isValueInRange!(short, int)(sht);
        else if (ti2 is typeid(uint))
            return isValueInRange!(short, uint)(sht); 
        else if (ti2 is typeid(long))
            return isValueInRange!(short, long)(sht);
        else if (ti2 is typeid(ulong))
            return isValueInRange!(short, ulong)(sht);
        else if (ti2 is typeid(float)) 
            return isValueInRange!(short, float)(sht);
        else if (ti2 is typeid(double))
            return isValueInRange!(short, double)(sht); 
        else if (ti2 is typeid(real))
            return isValueInRange!(short, real)(sht);
        else if (ti2 is typeid(ifloat))
            return isValueInRange!(short, ifloat)(sht);
        else if (ti2 is typeid(idouble))
            return isValueInRange!(short, idouble)(sht); 
        else if (ti2 is typeid(ireal))
            return isValueInRange!(short, ireal)(sht);    	        
    }    
    else if (ti1 is typeid(ushort))     
    {
        ushort usht =(va_arg!(ushort)(_argptr)); 
        if (ti2 is typeid(bool))
            return isValueInRange!(ushort, bool)(usht);
        else if (ti2 is typeid(byte))
            return isValueInRange!(ushort, byte)(usht);
        else if (ti2 is typeid(ubyte))
            return isValueInRange!(ushort, ubyte)(usht); 
        else if(ti2 is typeid(short))
            return isValueInRange!(ushort, short)(usht);
        else if (ti2 is typeid(ushort))
            return isValueInRange!(ushort, ushort)(usht); 
        else if(ti2 is typeid(int))
            return isValueInRange!(ushort, int)(usht);
        else if (ti2 is typeid(uint))
            return isValueInRange!(ushort, uint)(usht); 
        else if (ti2 is typeid(long))
            return isValueInRange!(ushort, long)(usht);
        else if (ti2 is typeid(ulong))
            return isValueInRange!(ushort, ulong)(usht);
        else if (ti2 is typeid(float)) 
            return isValueInRange!(ushort, float)(usht);
        else if (ti2 is typeid(double))
            return isValueInRange!(ushort, double)(usht); 
        else if (ti2 is typeid(real))
            return isValueInRange!(ushort, real)(usht);
        else if (ti2 is typeid(ifloat))
            return isValueInRange!(ushort, ifloat)(usht);
        else if (ti2 is typeid(idouble))
            return isValueInRange!(ushort, idouble)(usht); 
        else if (ti2 is typeid(ireal))
            return isValueInRange!(ushort, ireal)(usht);    	        
    }    
    else if (ti1 is typeid(int)) 
    {    
        int i = (va_arg!(int)(_argptr)); 
        if (ti2 is typeid(bool))
            return isValueInRange!(int, bool)(i);
        else if (ti2 is typeid(byte))
            return isValueInRange!(int, byte)(i);
        else if (ti2 is typeid(ubyte))
            return isValueInRange!(int, ubyte)(i); 
        else if(ti2 is typeid(short))
            return isValueInRange!(int, short)(i);
        else if (ti2 is typeid(ushort))
            return isValueInRange!(int, ushort)(i); 
        else if(ti2 is typeid(int))
            return isValueInRange!(int, int)(i);
        else if (ti2 is typeid(uint))
            return isValueInRange!(int, uint)(i); 
        else if (ti2 is typeid(long))
            return isValueInRange!(int, long)(i);
        else if (ti2 is typeid(ulong))
            return isValueInRange!(int, ulong)(i);
        else if (ti2 is typeid(float)) 
            return isValueInRange!(int, float)(i);
        else if (ti2 is typeid(double))
            return isValueInRange!(int, double)(i); 
        else if (ti2 is typeid(real))
            return isValueInRange!(int, real)(i);
        else if (ti2 is typeid(ifloat))
            return isValueInRange!(int, ifloat)(i);
        else if (ti2 is typeid(idouble))
            return isValueInRange!(int, idouble)(i); 
        else if (ti2 is typeid(ireal))
            return isValueInRange!(int, ireal)(i);    	
    }    
    else if (ti1 is typeid(uint))     
    {
        uint ui = (va_arg!(uint)(_argptr)); 
        if (ti2 is typeid(bool))
            return isValueInRange!(uint, bool)(ui);
        else if (ti2 is typeid(byte))
            return isValueInRange!(uint, byte)(ui);
        else if (ti2 is typeid(ubyte))
            return isValueInRange!(uint, ubyte)(ui); 
        else if(ti2 is typeid(short))
            return isValueInRange!(uint, short)(ui);
        else if (ti2 is typeid(ushort))
            return isValueInRange!(uint, ushort)(ui); 
        else if(ti2 is typeid(int))
            return isValueInRange!(uint, int)(ui);
        else if (ti2 is typeid(uint))
            return isValueInRange!(uint, uint)(ui); 
        else if (ti2 is typeid(long))
            return isValueInRange!(uint, long)(ui);
        else if (ti2 is typeid(ulong))
            return isValueInRange!(uint, ulong)(ui);
        else if (ti2 is typeid(float)) 
            return isValueInRange!(uint, float)(ui);
        else if (ti2 is typeid(double))
            return isValueInRange!(uint, double)(ui); 
        else if (ti2 is typeid(real))
            return isValueInRange!(uint, real)(ui);
        else if (ti2 is typeid(ifloat))
            return isValueInRange!(uint, ifloat)(ui);
        else if (ti2 is typeid(idouble))
            return isValueInRange!(uint, idouble)(ui); 
        else if (ti2 is typeid(ireal))
            return isValueInRange!(uint, ireal)(ui);            
    }    
    else if (ti1 is typeid(long))
    {     
        long l = (va_arg!(long)(_argptr));
        if (ti2 is typeid(bool))
            return isValueInRange!(long, bool)(l);
        else if (ti2 is typeid(byte))
            return isValueInRange!(long, byte)(l);
        else if (ti2 is typeid(ubyte))
            return isValueInRange!(long, ubyte)(l); 
        else if(ti2 is typeid(short))
            return isValueInRange!(long, short)(l);
        else if (ti2 is typeid(ushort))
            return isValueInRange!(long, ushort)(l); 
        else if(ti2 is typeid(int))
            return isValueInRange!(long, int)(l);
        else if (ti2 is typeid(uint))
            return isValueInRange!(long, uint)(l); 
        else if (ti2 is typeid(long))
            return isValueInRange!(long, long)(l);
        else if (ti2 is typeid(ulong))
            return isValueInRange!(long, ulong)(l);
        else if (ti2 is typeid(float)) 
            return isValueInRange!(long, float)(l);
        else if (ti2 is typeid(double))
            return isValueInRange!(long, double)(l); 
        else if (ti2 is typeid(real))
            return isValueInRange!(long, real)(l);
        else if (ti2 is typeid(ifloat))
            return isValueInRange!(long, ifloat)(l);
        else if (ti2 is typeid(idouble))
            return isValueInRange!(long, idouble)(l); 
        else if (ti2 is typeid(ireal))
            return isValueInRange!(long, ireal)(l);    	         
    }     
    else if (ti1 is typeid(ulong))   
    {  
        ulong ul = (va_arg!(ulong)(_argptr)); 
        if (ti2 is typeid(bool))
            return isValueInRange!(ulong, bool)(ul);
        else if (ti2 is typeid(byte))
            return isValueInRange!(ulong, byte)(ul);
        else if (ti2 is typeid(ubyte))
            return isValueInRange!(ulong, ubyte)(ul); 
        else if(ti2 is typeid(short))
            return isValueInRange!(ulong, short)(ul);
        else if (ti2 is typeid(ushort))
            return isValueInRange!(ulong, ushort)(ul); 
        else if(ti2 is typeid(int))
            return isValueInRange!(ulong, int)(ul);
        else if (ti2 is typeid(uint))
            return isValueInRange!(ulong, uint)(ul); 
        else if (ti2 is typeid(long))
            return isValueInRange!(ulong, long)(ul);
        else if (ti2 is typeid(ulong))
            return isValueInRange!(ulong, ulong)(ul);
        else if (ti2 is typeid(float)) 
            return isValueInRange!(ulong, float)(ul);
        else if (ti2 is typeid(double))
            return isValueInRange!(ulong, double)(ul); 
        else if (ti2 is typeid(real))
            return isValueInRange!(ulong, real)(ul);
        else if (ti2 is typeid(ifloat))
            return isValueInRange!(ulong, ifloat)(ul);
        else if (ti2 is typeid(idouble))
            return isValueInRange!(ulong, idouble)(ul); 
        else if (ti2 is typeid(ireal))
            return isValueInRange!(ulong, ireal)(ul);           
    }    
    else if (ti1 is typeid(float))  
    {   
        float f = (va_arg!(float)(_argptr)); 
        if (ti2 is typeid(bool))
            return isValueInRange!(float, bool)(f);
        else if (ti2 is typeid(byte))
            return isValueInRange!(float, byte)(f);
        else if (ti2 is typeid(ubyte))
            return isValueInRange!(float, ubyte)(f); 
        else if(ti2 is typeid(short))
            return isValueInRange!(float, short)(f);
        else if (ti2 is typeid(ushort))
            return isValueInRange!(float, ushort)(f); 
        else if(ti2 is typeid(int))
            return isValueInRange!(float, int)(f);
        else if (ti2 is typeid(uint))
            return isValueInRange!(float, uint)(f); 
        else if (ti2 is typeid(long))
            return isValueInRange!(float, long)(f);
        else if (ti2 is typeid(ulong))
            return isValueInRange!(float, ulong)(f);
        else if (ti2 is typeid(float)) 
            return isValueInRange!(float, float)(f);
        else if (ti2 is typeid(double))
            return isValueInRange!(float, double)(f); 
        else if (ti2 is typeid(real))
            return isValueInRange!(float, real)(f);
        else if (ti2 is typeid(ifloat))
            return isValueInRange!(float, ifloat)(f);
        else if (ti2 is typeid(idouble))
            return isValueInRange!(float, idouble)(f); 
        else if (ti2 is typeid(ireal))
            return isValueInRange!(float, ireal)(f);    	   
    }    
    else if (ti1 is typeid(double))  
    {   
        double d = (va_arg!(double)(_argptr));
        if (ti2 is typeid(bool))
            return isValueInRange!(double, bool)(d);
        else if (ti2 is typeid(byte))
            return isValueInRange!(double, byte)(d);
        else if (ti2 is typeid(ubyte))
            return isValueInRange!(double, ubyte)(d); 
        else if(ti2 is typeid(short))
            return isValueInRange!(double, short)(d);
        else if (ti2 is typeid(ushort))
            return isValueInRange!(double, ushort)(d); 
        else if(ti2 is typeid(int))
            return isValueInRange!(double, int)(d);
        else if (ti2 is typeid(uint))
            return isValueInRange!(double, uint)(d); 
        else if (ti2 is typeid(long))
            return isValueInRange!(double, long)(d);
        else if (ti2 is typeid(ulong))
            return isValueInRange!(double, ulong)(d);
        else if (ti2 is typeid(float)) 
            return isValueInRange!(double, float)(d);
        else if (ti2 is typeid(double))
            return isValueInRange!(double, double)(d); 
        else if (ti2 is typeid(real))
            return isValueInRange!(double, real)(d);
        else if (ti2 is typeid(ifloat))
            return isValueInRange!(double, ifloat)(d);
        else if (ti2 is typeid(idouble))
            return isValueInRange!(double, idouble)(d); 
        else if (ti2 is typeid(ireal))
            return isValueInRange!(double, ireal)(d);    	        
    }    
    else if (ti1 is typeid( real))  
    {   
        real r = (va_arg!(real)(_argptr)); 
        if (ti2 is typeid(bool))
            return isValueInRange!(real, bool)(r);
        else if (ti2 is typeid(byte))
            return isValueInRange!(real, byte)(r);
        else if (ti2 is typeid(ubyte))
            return isValueInRange!(real, ubyte)(r); 
        else if(ti2 is typeid(short))
            return isValueInRange!(real, short)(r);
        else if (ti2 is typeid(ushort))
            return isValueInRange!(real, ushort)(r); 
        else if(ti2 is typeid(int))
            return isValueInRange!(real, int)(r);
        else if (ti2 is typeid(uint))
            return isValueInRange!(real, uint)(r); 
        else if (ti2 is typeid(long))
            return isValueInRange!(real, long)(r);
        else if (ti2 is typeid(ulong))
            return isValueInRange!(real, ulong)(r);
        else if (ti2 is typeid(float)) 
            return isValueInRange!(real, float)(r);
        else if (ti2 is typeid(double))
            return isValueInRange!(real, double)(r); 
        else if (ti2 is typeid(real))
            return isValueInRange!(real, real)(r);
        else if (ti2 is typeid(ifloat))
            return isValueInRange!(real, ifloat)(r);
        else if (ti2 is typeid(idouble))
            return isValueInRange!(real, idouble)(r); 
        else if (ti2 is typeid(ireal))
            return isValueInRange!(real, ireal)(r);           
    }    
    else if (ti1 is typeid(ifloat))    
    { 
        ifloat ift = (va_arg!(ifloat)(_argptr)); 
        if (ti2 is typeid(bool))
            return isValueInRange!(ifloat, bool)(ift);
        else if (ti2 is typeid(byte))
            return isValueInRange!(ifloat, byte)(ift);
        else if (ti2 is typeid(ubyte))
            return isValueInRange!(ifloat, ubyte)(ift); 
        else if(ti2 is typeid(short))
            return isValueInRange!(ifloat, short)(ift);
        else if (ti2 is typeid(ushort))
            return isValueInRange!(ifloat, ushort)(ift); 
        else if(ti2 is typeid(int))
            return isValueInRange!(ifloat, int)(ift);
        else if (ti2 is typeid(uint))
            return isValueInRange!(ifloat, uint)(ift); 
        else if (ti2 is typeid(long))
            return isValueInRange!(ifloat, long)(ift);
        else if (ti2 is typeid(ulong))
            return isValueInRange!(ifloat, ulong)(ift);
        else if (ti2 is typeid(float)) 
            return isValueInRange!(ifloat, float)(ift);
        else if (ti2 is typeid(double))
            return isValueInRange!(ifloat, double)(ift); 
        else if (ti2 is typeid(real))
            return isValueInRange!(ifloat, real)(ift);
        else if (ti2 is typeid(ifloat))
            return isValueInRange!(ifloat, ifloat)(ift);
        else if (ti2 is typeid(idouble))
            return isValueInRange!(ifloat, idouble)(ift); 
        else if (ti2 is typeid(ireal))
            return isValueInRange!(ifloat, ireal)(ift);    	        
    }    
    else if (ti1 is typeid(idouble))  
    {   
        idouble id = (va_arg!(idouble)(_argptr)); 
        if (ti2 is typeid(bool))
            return isValueInRange!(idouble, bool)(id);
        else if (ti2 is typeid(byte))
            return isValueInRange!(idouble, byte)(id);
        else if (ti2 is typeid(ubyte))
            return isValueInRange!(idouble, ubyte)(id); 
        else if(ti2 is typeid(short))
            return isValueInRange!(idouble, short)(id);
        else if (ti2 is typeid(ushort))
            return isValueInRange!(idouble, ushort)(id); 
        else if(ti2 is typeid(int))
            return isValueInRange!(idouble, int)(id);
        else if (ti2 is typeid(uint))
            return isValueInRange!(idouble, uint)(id); 
        else if (ti2 is typeid(long))
            return isValueInRange!(idouble, long)(id);
        else if (ti2 is typeid(ulong))
            return isValueInRange!(idouble, ulong)(id);
        else if (ti2 is typeid(float)) 
            return isValueInRange!(idouble, float)(id);
        else if (ti2 is typeid(double))
            return isValueInRange!(idouble, double)(id); 
        else if (ti2 is typeid(real))
            return isValueInRange!(idouble, real)(id);
        else if (ti2 is typeid(ifloat))
            return isValueInRange!(idouble, ifloat)(id);
        else if (ti2 is typeid(idouble))
            return isValueInRange!(idouble, idouble)(id); 
        else if (ti2 is typeid(ireal))
            return isValueInRange!(idouble, ireal)(id);    	          
    }    
    else if (ti1 is typeid(ireal))  
    {   
        ireal ir = (va_arg!(ireal)(_argptr)); 
        if (ti2 is typeid(bool))
            return isValueInRange!(ireal, bool)(ir);
        else if (ti2 is typeid(byte))
            return isValueInRange!(ireal, byte)(ir);
        else if (ti2 is typeid(ubyte))
            return isValueInRange!(ireal, ubyte)(ir); 
        else if(ti2 is typeid(short))
            return isValueInRange!(ireal, short)(ir);
        else if (ti2 is typeid(ushort))
            return isValueInRange!(ireal, ushort)(ir); 
        else if(ti2 is typeid(int))
            return isValueInRange!(ireal, int)(ir);
        else if (ti2 is typeid(uint))
            return isValueInRange!(ireal, uint)(ir); 
        else if (ti2 is typeid(long))
            return isValueInRange!(ireal, long)(ir);
        else if (ti2 is typeid(ulong))
            return isValueInRange!(ireal, ulong)(ir);
        else if (ti2 is typeid(float)) 
            return isValueInRange!(ireal, float)(ir);
        else if (ti2 is typeid(double))
            return isValueInRange!(ireal, double)(ir); 
        else if (ti2 is typeid(real))
            return isValueInRange!(ireal, real)(ir);
        else if (ti2 is typeid(ifloat))
            return isValueInRange!(ireal, ifloat)(ir);
        else if (ti2 is typeid(idouble))
            return isValueInRange!(ireal, idouble)(ir); 
        else if (ti2 is typeid(ireal))
            return isValueInRange!(ireal, ireal)(ir);    	          
    }    
    
    return false;
}

//throw new Exception( "isNumeric(...) - Unsupported datatypes passed in."c );
debug( inrange )
{
int main()
{
	writefln();
    writefln("Compiled and tested with: \"%s v%d.%03d\"", scomp.name, scomp.version_major, scomp.version_minor);
    writefln("inrange.unittest finished!"c); 
	
    return 0;
}
}

unittest
{
    //writefln("** Started: Unittest general rules for datatypes. **");
    assert(isWholeNumber(bool.min) == true);
    assert(isWholeNumber(bool.max) == true);	
    assert(isNegative(bool.min) == false);
    assert(isNegative(bool.max) == false);	
    
    assert(isWholeNumber(byte.min) == true);
    assert(isWholeNumber(byte.max) == true);	
    assert(isNegative(byte.min) == true);
    assert(isNegative(byte.max) == false);    
    
    assert(isWholeNumber(ubyte.min) == true);
    assert(isWholeNumber(ubyte.max) == true);	
    assert(isNegative(ubyte.min) == false);
    assert(isNegative(ubyte.max) == false);    

    assert(isWholeNumber(short.min) == true);
    assert(isWholeNumber(short.max) == true);	
    assert(isNegative(short.min) == true);
    assert(isNegative(short.max) == false);    
    
    assert(isWholeNumber(ushort.min) == true);
    assert(isWholeNumber(ushort.max) == true);	
    assert(isNegative(ushort.min) == false);
    assert(isNegative(ushort.max) == false); 

    assert(isWholeNumber(int.min) == true);
    assert(isWholeNumber(int.max) == true);	
    assert(isNegative(int.min) == true);
    assert(isNegative(int.max) == false);    
    
    assert(isWholeNumber(uint.min) == true);
    assert(isWholeNumber(uint.max) == true);	
    assert(isNegative(uint.min) == false);
    assert(isNegative(uint.max) == false);     

    assert(isWholeNumber(long.min) == true);
    assert(isWholeNumber(long.max) == true);	
    assert(isNegative(long.min) == true);
    assert(isNegative(long.max) == false);    
    
    assert(isWholeNumber(ulong.min) == true);
    assert(isWholeNumber(ulong.max) == true);	
    assert(isNegative(ulong.min) == false);
    assert(isNegative(ulong.max) == false);   

    //writefln("float.min=%g, float.max=%g", float.min, float.max);
    assert(isWholeNumber(float.min) == false);
    assert(isWholeNumber(float.max) == true);	
    assert(isNegative(float.min) == false);
    assert(isNegative(float.max) == false);    
    
    assert(isWholeNumber(double.min) == false);
    assert(isWholeNumber(double.max) == true);	
    assert(isNegative(double.min) == false);
    assert(isNegative(double.max) == false);   

    assert(isWholeNumber(real.min) == false);
    assert(isWholeNumber(real.max) == true);	
    assert(isNegative(real.min) == false);
    assert(isNegative(real.max) == false); 

    assert(isWholeNumber(ifloat.min) == false);
    assert(isWholeNumber(ifloat.max) == true);	
    assert(isNegative(ifloat.min) == false);
    assert(isNegative(ifloat.max) == false);    
    
    assert(isWholeNumber(idouble.min) == false);
    assert(isWholeNumber(idouble.max) == true);	
    assert(isNegative(idouble.min) == false);
    assert(isNegative(idouble.max) == false);   

    assert(isWholeNumber(ireal.min) == false);
    assert(isWholeNumber(ireal.max) == true);	
    assert(isNegative(ireal.min) == false);
    assert(isNegative(ireal.max) == false);    
    
    assert(isWholeNumber(-0.0) == true);
    assert(isNegative(-0.0) == true);            
    
    assert(isWholeNumber(-float.nan) == false);
    assert(isNegative(-float.nan) == true);
    
    //writefln("float.infinity=%f", float.infinity);
    assert(isWholeNumber(float.infinity) == true);                          
    assert(isNegative(float.infinity) == false); 
    assert(isWholeNumber(-float.infinity) == true);                          
    assert(isNegative(-float.infinity) == true); 
    
    assert(isWholeNumber(-double.nan) == false);
    assert(isNegative(-double.nan) == true); 
    assert(isWholeNumber(double.infinity) == true);                          
    assert(isNegative(double.infinity) == false); 
    assert(isWholeNumber(-double.infinity) == true);                          
    assert(isNegative(-double.infinity) == true); 
        
    assert(isWholeNumber(-real.nan) == false);
    assert(isNegative(-real.nan) == true);  
    assert(isWholeNumber(real.infinity) == true);                          
    assert(isNegative(real.infinity) == false); 
    assert(isWholeNumber(-real.infinity) == true);                          
    assert(isNegative(-real.infinity) == true); 
                
 /+   
             writefln("T1.typid=%s, T1.v=", typeid(T1), v,
                      " T1.tsize=%d, isNeg=%b, isWhole=%b ", 
                       T1.sizeof, isNegative(v), isWholeNumber(v), 
                      "T2.typid=%s, T2.tsize=%d", typeid(T2), T2.sizeof); 
+/	
    //writefln("** Ended: Unittest general rules for datatypes. **");
    //writefln();
}

unittest
{
    debug( inrange ) writefln("Unittest started"c);
    ulong ul = 34567890123UL;

    assert(isNumValueInRange(-123, TF_ULONG) == false);     // IsNegative
    assert(isNumValueInRange(long.min, TF_ULONG) == false); // IsNegative
    assert(isNumValueInRange(123UL, TF_ULONG) == true);     // Same datatype
    assert(isNumValueInRange(ulong.max, TF_UINT) == false); // too large
    
    Object o = new Object;
    assert(isNumValueInRange(ulong.max, o) == false);       // object is not a number
    assert(isNumValueInRange(o, TF_UINT) == false);         // object is not a number
    
    cfloat cf;
    assert(isNumValueInRange(ulong.min, cf) == false);      // complex-float return false
    assert(isNumValueInRange(cf, TF_ULONG) == false);       // complex-float return false
   
    assert(isNumValueInRange(ul, TF_UINT) == false);        // too large
    assert(isNumValueInRange(123.89L, TF_REAL) == true);    // Same range
    assert(isNumValueInRange(456U, TF_LONG) == true);       // In range
    assert(isNumValueInRange(uint.max, TF_LONG) == true);   // In range
    assert(isNumValueInRange(ulong.max, TF_LONG) == false); // too large
    assert(isNumValueInRange(long.min, TF_INT) == false);   // too small
    assert(isNumValueInRange(ulong.max, TF_UINT) == false); // too large
    assert(isNumValueInRange(int.min, TF_LONG) == true);    // In range
    assert(isNumValueInRange(int.max, TF_LONG) == true);    // In range
    assert(isNumValueInRange(true, TF_LONG) == true);       // In range
    assert(isNumValueInRange(2L, TF_BOOL) == false);        // too large
    assert(isNumValueInRange(-101L, TF_BOOL) == false);     // too small
    assert(isNumValueInRange(127, TF_USHORT) == true);      // In range
    assert(isNumValueInRange(345.0f, TF_INT) == true); 
    assert(isNumValueInRange(345.95f, TF_INT) == false); 
    assert(isNumValueInRange(345.0E+0f, TF_INT) == true); 
    
    assert(isNumValueInRange(345, TF_FLOAT) == true); 
    assert(isNumValueInRange(345U, TF_FLOAT) == true); 
    assert(isNumValueInRange(cast(short)345, TF_FLOAT) == true); 
    assert(isNumValueInRange(cast(ushort)345, TF_FLOAT) == true); 
    assert(isNumValueInRange(cast(byte)-127, TF_FLOAT) == true); 
    assert(isNumValueInRange(cast(byte)-0, TF_FLOAT) == true); 
    assert(isNumValueInRange(cast(ubyte)127, TF_FLOAT) == true); 
    assert(isNumValueInRange(true, TF_FLOAT) == true); 
    assert(isNumValueInRange(345L, TF_FLOAT) == true); 
    assert(isNumValueInRange(345UL, TF_FLOAT) == true); 

    assert(isNumValueInRange(345.0, TF_INT) == true); 
    assert(isNumValueInRange(345.95, TF_INT) == false);
    assert(isNumValueInRange(345.0E+0, TF_INT) == true);
    assert(isNumValueInRange(345, TF_DOUBLE) == true);
    assert(isNumValueInRange(345U, TF_DOUBLE) == true);
    assert(isNumValueInRange(cast(short)345, TF_DOUBLE) == true);
    assert(isNumValueInRange(cast(ushort)345U, TF_DOUBLE) == true);
    assert(isNumValueInRange(cast(byte)-127, TF_DOUBLE) == true);
    assert(isNumValueInRange(cast(byte)-0, TF_DOUBLE) == true);
    assert(isNumValueInRange(cast(ubyte)127U, TF_DOUBLE) == true);
    assert(isNumValueInRange(true, TF_DOUBLE) == true);
    assert(isNumValueInRange(345L, TF_DOUBLE) == true);
    assert(isNumValueInRange(345UL, TF_DOUBLE) == true);
      
    assert(isNumValueInRange(345.0L, TF_INT) == true);
    assert(isNumValueInRange(345.95L, TF_INT) == false);
    assert(isNumValueInRange(345.0E+0L, TF_INT) == true);
    assert(isNumValueInRange(345, TF_REAL) == true);
    assert(isNumValueInRange(345U, TF_REAL) == true);
    assert(isNumValueInRange(cast(short)345, TF_REAL) == true);
    assert(isNumValueInRange(cast(ushort)345U, TF_REAL) == true);
    assert(isNumValueInRange(cast(byte)-127, TF_REAL) == true);
    assert(isNumValueInRange(cast(byte)-0, TF_REAL) == true);
    assert(isNumValueInRange(cast(ubyte)127U, TF_REAL) == true);
    assert(isNumValueInRange(true, TF_REAL) == true);
    assert(isNumValueInRange(long.max, TF_REAL) == true);
    assert(isNumValueInRange(long.min, TF_REAL) == true);
    assert(isNumValueInRange(345UL, TF_REAL) == true);
 
    // Imaginary Floats
    assert(isNumValueInRange(345.0fi, TF_INT) == true);
    assert(isNumValueInRange(345.95fi, TF_INT) == false);
    assert(isNumValueInRange(345.0E+0fi, TF_INT) == true);
    assert(isNumValueInRange(345.0i, TF_INT) == true);
    assert(isNumValueInRange(345.95i, TF_INT) == false);
    assert(isNumValueInRange(345.0E+0i, TF_INT) == true);
    assert(isNumValueInRange(345.0Li, TF_INT) == true);
    assert(isNumValueInRange(345.95Li, TF_INT) == false);
    assert(isNumValueInRange(345.0E+0Li, TF_INT) == true);
     
    assert(isNumValueInRange(345, TF_IFLOAT) == true);
    assert(isNumValueInRange(int.min, TF_IFLOAT) == true);
    assert(isNumValueInRange(int.max, TF_IFLOAT) == false);
    
    ifloat if2 = long.min * -1.0i;
    //writefln("long.min=%d, ifloat=%g ", long.min, if2);
//    assert(isNumValueInRange(long.min, TF_IFLOAT) == false);
//    assert(isNumValueInRange(long.max, TF_IFLOAT) == false);
    assert(isNumValueInRange(ulong.min, TF_IFLOAT) == true);
//    assert(isNumValueInRange(ulong.min, TF_IFLOAT) == false);
    
    assert(isNumValueInRange(uint.min, TF_IFLOAT) == true);
    assert(isNumValueInRange(uint.max, TF_IFLOAT) == false);
    assert(isNumValueInRange(345, TF_IDOUBLE) == true);
    assert(isNumValueInRange(uint.min, TF_IDOUBLE) == true);
    assert(isNumValueInRange(uint.max, TF_IDOUBLE) == true);
    assert(isNumValueInRange(345, TF_IREAL) == true);
    assert(isNumValueInRange(uint.min, TF_IREAL) == true);
    assert(isNumValueInRange(uint.max, TF_IREAL) == true);
}