/+
 ' Source          : isNumeric.d
 ' Version         : v1.0
 ' Author          : David L. 'SpottedTiger' Davis
 ' Source Platform : Intel x86 / WinXP SP2 
 ' Target Platform : Should work on any CPU / OS combo that D compiles on.
 '                 :
 ' Date Created    : 31.May.05 Compiled and Tested with dmd v0.125
 ' Date Modified   : 01.Jun.05 Modified the function to handle the
 '                 :           imaginary and complex float-point 
 '                 :           datatypes.
 '                 : 02.Jun.05 Added support for void[] to isNumeric(...)
 '                 :           function since can accept char[] input  
 '                 :           directly and therefore could be used to 
 '                 :           store a numeric string.
 '                 : 07.Jun.05 Added a version of the isNumeric() function
 '                 :           that return the flags of the type of
 '                 :           numeric found in the string; which was
 '                 :           very useful in the toNumeric() function.
 '                 : 09.Jun.05 Recompiled and unittested this code under D v0.126.
 '                 : 26.Aug.05 Did some minor code cleanup.
 '                 : 16.Sep.05 Added a check for an Imaginary Infinity like
 '                 :           "infi" and -infi" which may be added in the
 '                 :           future, and I want to be ready for it. 
 '                 : 24.Oct.05 Compiled and tested with dmd v0.137.
 '                 : 30.Dec.05 Compiled and tested with dmd v0.142.
 '                 : 23.Mar.06 Compiled and tested with dmd v0.150.
 '                 : 28.May.06 Recompiled and Tested with dmd v0.159.
 '                 : 19.Jul.06 All "Imports" are now private by default (which 
 '                 :           shouldn't be a problem, since I always use the 
 '                 :           private keyword) in dmd v0.163, recompiled and tested.
 '                 : 09.Sep.06 Complied and tested with dmd v0.166.
 '                 : 14.Sep.06 In char[], wchar[], and dchar[] datatype be sure to
 '                 :           pass the optional "bool bAllowSep" parameter.
 '                 : 23.Oct.06 Complied and tested with dmd v0.172.
 '                 : 05.Nov.06 Compiled and tested with dmd v0.173.
 '                 : 09.Dec.06 Compiled and tested with dmd v0.177.
 '                 : 05.Jan.07 Compiled and tested with dmd v1.0.
 '                 : 09.Feb.07 Compiled and tested with dmd v1.005. 
 '                 : 03.Jan.08 Compiled and tested with dmd v1.025.
 '                 : 20.Mar.08 Compiled and tested with dmd v1.028. 
 '                 : 19.Jun.08 Compiled and tested with dmd v1.031.
 '                 : 17.Jul.08 Compiled and tested with dmd v1.033. 
 '                 : 13.May.09 Compiled and tested with dmd v1.045.
 '                 :
 ' Licence         : Public Domain / Contributed to Digital Mars
 ' --------------------------------------------------------------
 ' To Unittest: dmd isnumeric.d -debug=isnumeric -unittest
 '
 ' If using isnumeric.d alone in another project
 ' To Complie : dmd mycode.d isnumeric.d
 +/
module isnumeric;

debug( isnumeric ) private import scomp = std.compiler;
private import std.stdio;  // for writefln();
private import std.string; // for tolower()
private import std.stdarg; // for argument gathering
private import std.utf;    // for toUTF8()

/+
 ' final bool isNumeric(in char[], out bool[], in bool = false)
 ' ---------------------------------------------------------------
 ' [in] char[] s can be formatted in the following ways:
 '
 ' Integer Whole Number:
 ' (for byte, ubyte, short, ushort, int, uint, long, and ulong)
 ' ['+'|'-']digit(s)[U|L|UL]
 '
 ' examples: 123, 123UL, 123L, +123U, -123L
 '
 ' Floating-Point Number:
 ' (for float, double, real, ifloat, idouble, and ireal)
 ' ['+'|'-']digit(s)[.][digit(s)][[e-|e+]digit(s)][i|f|L|Li|fi]]
 '      or [nan|nani|inf|-inf]
 '
 ' examples: +123., -123.01, 123.3e-10f, 123.3e-10fi, 123.3e-10L
 ' 
 ' (for cfloat, cdouble, and creal)
 ' ['+'|'-']digit(s)[.][digit(s)][[e-|e+]digit(s)][+]
 '         [digit(s)[.][digit(s)][[e-|e+]digit(s)][i|f|L|Li|fi]]
 '      or [nan|nani|nan+nani|inf|-inf]
 '
 ' examples: nan, -123e-1+456.9e-10Li, +123e+10+456i, 123+456i
 '
 ' [out] bool[] bflags
 ' Returns four flag values that are helpful in selecting the lowest 
 ' possible numeric datatype that the value will fit in, which will 
 ' also conserve memory.  
 '
 ' These four flags are as follows:
 ' ISINTEGER   = 0 : if true, it indicates a Integer Whole Number.
 ' ISNEGATIVE  = 1 : if true, it indicates a Negative Integer or Floating-Point Number.
 ' ISIMAGINARY = 2 : if true, it indicates a Imaginary Floating-Point Number.
 ' ISCOMPLEX   = 3 : if true, it indicates a Complex Floating-Point Number.
 '
 ' [in] bool bAllowSep 
 ' False by default, but when set to true it will accept the 
 ' separator characters "," and "_" within the string, but these  
 ' characters should be stripped from the string before using any 
 ' of the conversion functions like toInt(), toFloat(), and etc 
 ' else an error will occur.
 '
 ' Also please note, that no spaces are allowed within the string  
 ' anywhere whether it's a leading, trailing, or embedded space(s), 
 ' thus they too must be stripped from the string before using this
 ' function, or any of the conversion functions.
 +/
bool isNumeric(in char[] s, out bool[] bflags, in bool bAllowSep = false)
{
    int    iLen = s.length;
    bool   bDecimalPoint = false;
    bool   bExponent = false;
    bool   bComplex = false;
    char[] sx = std.string.tolower(s); 
    int    j  = 0;
    char   c;

    enum
    {
        ISINTEGER   = 0,
        ISNEGATIVE  = 1,
        ISIMAGINARY = 2,
        ISCOMPLEX   = 3
    }

    void setFlags()
    {
        char c;
    
        // Set length to an even 8-bits / 1 byte
        if (bflags.length < 8)
            bflags.length = bflags.length + 8;
    
        // Set bInt (is a whole numer) value 
        if (!bDecimalPoint && !bExponent && !bComplex)
            bflags[ISINTEGER] = true;
        else
            // (is a floating-point number)
            bflags[ISINTEGER] = false;
        
        // Set bNeg (is a negative number) value     
        if (sx[0] == '-')        
            bflags[ISNEGATIVE] = true;
        else    
            bflags[ISNEGATIVE] = false;
            
        // Set bISuffix (is a imaginary floating-point number) value 
        c = sx[sx.length - 1];
        if (c == 'i')
            bflags[ISIMAGINARY] = true;
        else
            bflags[ISIMAGINARY] = false; 
        
        // Set bComplex (is a complex floating-point number) value       
        bflags[ISCOMPLEX] = bComplex;
    }
    
    //writefln("isNumeric(char[], bool = false) called!");
    // Empty string, return false
    if (iLen == 0)
        return false;
    
    // Check for NaN (Not a Number)
    if (sx == "nan"c || sx == "nani"c || 
        sx == "nan+nani"c || sx == "nan+0i"c)
    {
        setFlags();    
        return true;
    }    
        
    // Check for Infinity
    if (sx == "inf"c  || sx == "-inf"c ||
        sx == "infi"c || sx == "-infi"c)
    {
        setFlags();    
        return true;
    }
     
    // A sign is allowed only in the 1st character   
    if (sx[0] == '-' || sx[0] == '+')
        j++;
            
    for (int i = j; i < iLen; i++)
    {
        c = sx[i];
    
        // Digits are good, continue checking 
        // with the next character... ;)
        if (c >= '0' && c <= '9') 
        {
            continue;
        }
        // Check for the complex type, and if found 
        // reset the flags for checking the 2nd number.  
        else if (c == '+')
        {
            if (i > 0) 
            {
                bDecimalPoint = false;
                bExponent = false;
                bComplex = true;
                continue;
            }
            else
                return false;
        }        
        // Allow only one exponent per number   
        else if (c == 'e')  
        {
            // A 2nd exponent found, return not a valid number
            if (bExponent)
                return false;
                
            if (i + 1 < iLen)
            {
                // Look forward for the sign, and if 
                // missing then this is not a number.
                if (sx[i + 1] != '-' && sx[i + 1] != '+')
                    return false;
                else
                {
                    bExponent = true;
                    i++;    
                }    
            }        
            else
                // Ending in "E", return not a valid number
                return false;        
        }  
        // Allow only one decimal point per number to be used
        else if (c == '.')
        {
            // A 2nd decimal point found, return not a valid number
            if (bDecimalPoint)
                return false;
            
            bDecimalPoint = true;
            continue;
        }   
        // Check for ending literal characters: "f,u,l,i,ul,fi,li",
        // and wheater they're being used with the correct datatype.
        else if (i == iLen - 2)
        {
            // Integer Whole Number
            if (sx[i..iLen] == "ul"c && 
               (!bDecimalPoint && !bExponent && !bComplex))
            {   
                setFlags();
                return true;
            }    
            // Floating-Point Number
            else if ((sx[i..iLen] == "fi"c || sx[i..iLen] == "li"c) &&
                     (bDecimalPoint || bExponent || bComplex))
            {         
                setFlags();     
                return true;
            }    
            else if (sx[i..iLen] == "ul"c && 
                    (bDecimalPoint || bExponent || bComplex))
            {
                return false;    
            }    
            // Could be a Integer or a Float, thus
            // all these suffixes are valid for both  
            else if (sx[i..iLen] == "ul"c || 
                     sx[i..iLen] == "fi"c || 
                     sx[i..iLen] == "li"c)
            {         
                setFlags();     
                return true;
            }    
            else    
                return false;
        }
        else if (i == iLen - 1)
        {
            // Integer Whole Number
            if ((c == 'u' || c == 'l') && 
                (!bDecimalPoint && !bExponent && !bComplex))
            {
                setFlags();    
                return true;
            }    
            // Check to see if the last character in the string 
            // is the required 'i' character
            else if (bComplex)
            {
                if (c == 'i')
                {
                    setFlags();
                    return true;
                }    
                else 
                    return false;        
            }
            // Floating-Point Number
            else if ((c == 'l' || c == 'f' || c == 'i') &&
                     (bDecimalPoint || bExponent))
            {         
                setFlags();
                return true;
            }    
            // Could be a Integer or a Float, thus  
            // all these suffixes are valid for both 
            else if (c == 'l' || c == 'f' || c == 'i')
            {
                setFlags();
                return true;
            }    
            else
                return false;
        }
        else
            // Check if separators are allow  
            // to be in the numeric string
            if (bAllowSep == true && (c == '_' || c == ','))
                continue;
            else    
                return false;       
    }     
    
    setFlags();
    return true;
}

bool isNumeric(in char[] s, in bool bAllowSep = false)
{
    bool[] bflags;
    
    return isNumeric(s, bflags, bAllowSep);
}

// Allow any object as a parameter
bool isNumeric(...)
{
    return isNumeric(_arguments, _argptr);
}

// Check only the first parameter for the value, and the second is there is one and it's a bool 
// datatype use it for char[], wchar[], and dchar[], else all other parameters will be ignored. 
bool isNumeric(TypeInfo[] _arguments, va_list _argptr)
{
    TypeInfo ti = _arguments[0];
    bool     bChkAllowSep = false;
    
    //writefln("isNumeric(...) called!");
    if (!_arguments.length)
        return false;
    else if (_arguments.length >= 2)
    {
        if (_arguments[1] is typeid(bool))
            bChkAllowSep = true;
    }

    if (ti is typeid(char[]))
    {
        char[] s  = va_arg!(char[])(_argptr);
        if (bChkAllowSep)
        {
            bool bAllowSep = va_arg!(bool)(_argptr);
            return isNumeric(s, bAllowSep);
        }
        else
            return isNumeric(s, false);
    }    
    else if (ti is typeid(wchar[]))
    {
        char[] s = std.utf.toUTF8(va_arg!(wchar[])(_argptr));
        if (bChkAllowSep)
        {
            bool bAllowSep = va_arg!(bool)(_argptr);
            return isNumeric(s, bAllowSep);
        }
        else
            return isNumeric(s, false);
    }    
    else if (ti is typeid(dchar[]))
    {
        char[] s = std.utf.toUTF8(va_arg!(dchar[])(_argptr));
        if (bChkAllowSep)
        {
            bool bAllowSep = va_arg!(bool)(_argptr);
            return isNumeric(s, bAllowSep);
        }
        else
            return isNumeric(s, false);
    }    
    else if (ti is typeid(real))
        return true;
    else if (ti is typeid(double)) 
        return true;   
    else if (ti is typeid(float)) 
        return true;  
    else if (ti is typeid(ulong)) 
        return true; 
    else if (ti is typeid(long)) 
        return true;   
    else if (ti is typeid(uint)) 
        return true;  
    else if (ti is typeid(int)) 
        return true;   
    else if (ti is typeid(ushort)) 
        return true;   
    else if (ti is typeid(short)) 
        return true;   
    else if (ti is typeid(ubyte)) 
    {
       char[] s = ""c;
       s.length = 1;
       s[0]= va_arg!(ubyte)(_argptr);
       return isNumeric(cast(char[])s);
    }
    else if (ti is typeid(byte)) 
    {
       char[] s = ""c;
       s.length = 1;
       s[0] = va_arg!(byte)(_argptr);
       return isNumeric(cast(char[])s);
    }
    else if (ti is typeid(ireal))
        return true;
    else if (ti is typeid(idouble)) 
        return true;   
    else if (ti is typeid(ifloat)) 
        return true;  
    else if (ti is typeid(creal))
        return true;
    else if (ti is typeid(cdouble)) 
        return true;   
    else if (ti is typeid(cfloat)) 
        return true;  
    else if (ti is typeid(char))
    {
        char[] s = ""c;
        s.length = 1;
        s[0] = va_arg!(char)(_argptr);
        return isNumeric(s);
    }
    else if (ti is typeid(wchar))
    {
    	wchar[] ws = ""w;
        ws.length = 1;
        ws[0] = va_arg!(wchar)(_argptr);
        return isNumeric(std.utf.toUTF8(ws));
    }
    else if (ti is typeid(dchar))
    { 
        dchar[] ds = ""d;
        ds.length = 1;
        ds[0] = va_arg!(dchar)(_argptr);
        return isNumeric(std.utf.toUTF8(ds));
    }
    else if (ti is typeid(void[]))
        return isNumeric(std.utf.toUTF8(cast(char[])va_arg!(void[])(_argptr)));
    //else if (ti is typeid(cent)) 
    //    return true;   
    //else if (ti is typeid(ucent)) 
    //    return true;  
    else       
       throw new Exception( "isNumeric(...) - Unsupported datatypes passed in."c );
       //return false; 
}

unittest
{
    writefln( "isNumeric(in char[], bool = false).unittest"c );
    char[] s;
    void[] vo;
    
    // Test the isNumeric(in char[]) function
    assert(.isNumeric("1") == true );
    assert(.isNumeric("1.0") == true );
    assert(.isNumeric("1e-1") == true );
    assert(.isNumeric("12345xxxx890") == false );
    assert(.isNumeric("567L") == true );
    assert(.isNumeric("23UL") == true );
    assert(.isNumeric("-123..56f") == false );
    assert(.isNumeric("12.3.5.6") == false );
    assert(.isNumeric(" 12.356") == false );
    assert(.isNumeric("123 5.6") == false );
    assert(.isNumeric("1233E-1+1.0e-1i") == true );

    assert(.isNumeric("123.00E-5+1234.45E-12Li") == true);
    assert(.isNumeric("123.00e-5+1234.45E-12iL") == false);
    assert(.isNumeric("123.00e-5+1234.45e-12uL") == false);
    assert(.isNumeric("123.00E-5+1234.45e-12lu") == false);
  
    assert(.isNumeric("123fi") == true);
    assert(.isNumeric("123li") == true);
    assert(.isNumeric("--123L") == false);
    assert(.isNumeric("+123.5UL") == false);
    assert(.isNumeric("123f") == true);
    assert(.isNumeric("123.u") == false);

    assert(.isNumeric(std.string.toString(real.nan)) == true);
    assert(.isNumeric(std.string.toString(-real.infinity)) == true);
    assert(.isNumeric(std.string.toString(123e+2+1234.78Li)) == true);
 
    s = "$250.99-"c;
    assert(.isNumeric(s[1..s.length - 2]) == true);
    assert(.isNumeric(s) == false);
    assert(.isNumeric(s[0..s.length - 1]) == false);

    // These test calling the isNumeric(...) function
    assert(.isNumeric(1,123UL) == true);
    assert(.isNumeric('2') == true);
    assert(.isNumeric('x') == false);
    assert(.isNumeric(cast(byte)0x57) == false); // 'W'
    assert(.isNumeric(cast(byte)0x37) == true);  // '7'
    assert(.isNumeric(cast(wchar[])"145.67") == true);
    assert(.isNumeric(cast(dchar[])"145.67U") == false);
    assert(.isNumeric(123_000.23fi) == true);
    assert(.isNumeric(123.00E-5+1234.45E-12Li) == true);
    assert(.isNumeric(real.nan) == true);
    assert(.isNumeric(-real.infinity) == true);
   
    vo = "123.99E-5+1234.45E-12i"c;
    assert(.isNumeric(vo) == true);
}

debug( isnumeric )
{
int main()
{
	writefln();
    writefln("Compiled and tested with: \"%s v%d.%03d\"", scomp.name, scomp.version_major, scomp.version_minor);
    writefln("isnumeric.unittest finished!"c); 
	
    return 0;
}
}