/+
 ' Module          : MarsDate.d v1.0
 '                 : Uses Thomas Gangale's Darian System jscript code, 
 '                 : which was converted and modified to work in D.
 '                 :
 ' Author          : David L. 'SpottedTiger' Davis
 ' Source Platform : Intel x86 / WinXP SP2 
 ' Target Platform : Should work on any CPU / OS combo that D compiles on.
 '                 :
 ' Created Date    : 15.Jul.05 Compiled and tested with dmd v0.128.
 ' Modified Date   : 17.Jul.05 Corrected a number of issues.
 '                 : 12.Aug.05 Compiled and tested with dmd v0.129.
 '                 : 10.Sep.05 Added an updated version of stringbuffer.d
 '                 :           recompiled and tested with dmd v0.131.
 '                 : 26.Oct.05 Recompiled and tested with dmd v0.137.    
 '                 : 30.Dec.05 Recompiled and tested with dmd v0.142.  
 '                 : 22.Mar.06 Recompiled and tested with dmd v0.150.  
 '                 : 28.May.06 Recompiled and Tested with dmd v0.159.
 '                 : 19.Jul.06 All "Imports" are now private by default 
 '                 :           (which shouldn't be a problem, since I
 '                 :           always use the private keyword) in 
 '                 :           dmd v0.163, recompiled and tested.
 '                 : 23.Oct.06 Recompiled and Tested with dmd v0.172.
 '                 : 05.Nov.06 Recompiled and Tested with dmd v0.173.
 '                 : 09.Dec.06 Recompiled and Tested with dmd v0.177.
 '                 : 05.Jan.07 Recompiled and Tested with dmd v1.0.
 '                 : 09.Feb.07 Recompiled and Tested with dmd v1.005. 
 '                 : 03.Jan.08 Recompiled and Tested with dmd v1.025.
 '                 : 19.Jun.08 Recompiled and Tested with dmd v1.031. 
 '                 : 17.Jul.08 Recompiled and Tested with dmd v1.033. 
 '                 : 13.May.09 Recompiled and Tested with dmd v1.045.  
 '                 :
 ' Requirements    : std.math, std.date, std.string, and stringbuffer.d
 '                 :
 ' Licence         : Public Domain for this D module, but the 
 '                 : "Darian System" logic is copyrighted 
 '                 : (c)1986-2005 by Thomas Gangale.
 ' ---------------------------------------------------------------------------
 ' 
 ' To complie for a unittest:
 ' dmd MarsDate.d StringBuffer.d -debug=marsdate -unittest
 '
 ' To compile a D program that use MarsDate.d:
 ' dmd MySource.d MarsDate.d StringBuffer.d
 +/
module marsdate; 

debug( marsdate ) private import scomp = std.compiler; 
private import std.stdio;
private import std.math;
private import std.date;
private import std.string;
private import stringbuffer;

// to convert Earth time to Mars:
const long MILLISECONDS_PER_SOL    = 0x54A9A4C; 
const long MILLISECONDS_PER_HOUR   = 0x36EE80;
const long MILLISECONDS_PER_MINUTE = 0xEA60;
const long MILLISECONDS_PER_SECOND = 0x3E8;

// Viking 1 Lander: Landed on Mars on 20 July 1976 11:53:06 UTC 
// (4:13 p.m. local Mars time) at Chryse Planitia at (22.533 deg N, 48.264 deg W).
//
// start date for Viking 1
// (jscript) var MARS_START = Date.UTC(1975, 11, 17, 23, 46, 12, 0).valueOf();
// start date for Martian northern vernal equinox in 1609, the year
// Astronomy Novia was published by Johannes Kepler, and also the year
// the telescope was first used for astronomy, by Galileo Galilei:
// 1609 Mar 11 08:40:48
// (jscript) const MARS_START = Date.UTC(1609, 02, 11, 18, 40, 48, 400).valueOf();
const d_time MARS_START = 0xFFFFF5A4FE789D10; // -11385983951600;

struct marsdatetime
{
    int  iMir;
    int  iMonth;
    int  iSol;
    int  iSpells;
    int  iMoments;
    int  iJiffies;
    int  ims;
}

// returns true if a leap mir:
bool isLeapMir(in int iMir)
{
    // if mir is not divisible by 10 then the rule is very simple:
    // it's a leap year if odd, and not a leap year if even:
    if (iMir % 10 != 0)
        return (iMir % 2 == 1);

    // now we know mir is divisible by 10, must investigate further:
    // if mir divisible by 500, then it's a leap mir:
    if (iMir % 1000 == 0)
        return true;

    // if mir divisible by 100 then it's not a leap mir:
    if (iMir % 100 == 0)
        return false;

    // otherwise it is a leap mir:
    return true;
}

// returns the current position of sols within a specified mir
int getSolInMir(in d_time dtDate)
{
    marsdatetime mdtDate = getMarsDate(dtDate);
    int iSols = 0;

    for ( int i = 1; i < mdtDate.iMonth; i++ )
        iSols += getSolsInMonth(i, mdtDate.iMir);

    return mdtDate.iSol + iSols;
    
} // end getSolInMir()

// returns the total number of sols in the specified mir
int getSolsInMir(in int iMir)
{
    if (isLeapMir(iMir))
        return 669;
    else
        return 668;

} // end getSolsInMir()

// returns number of sols in month, specified by month and mir:
int getSolsInMonth(in int iMonth, in int iMir)
{
    // Vrishika is a special case:
    if (iMonth == 24)
    {
        if (isLeapMir(iMir))
            return 28;
        else
            return 27;
    }
    else if (iMonth == 6 || iMonth == 12 || iMonth == 18)
    {
        return 27;
    }
    else
        return 28;
}

// MarsDate Struct.
// input: time in milliseconds (standard return from (new Date()).valueOf())
// output: object representing Mars date and time
marsdatetime getMarsDate(in d_time dtDate = d_time_nan)
{
    marsdatetime mdtDate;
    long lMilliseconds = dtDate;

    real rSols     = 0.0L;
    real rTime     = 0.0L;
    int  iSpells   = 0;
    int  iMoments  = 0;
    int  iJiffies  = 0;
    int  ims       = 0;
    int  iMir      = 0;
    int  iMirLen   = 0;
    // start in Sagittarius (month 1):
    int  iMonth    = 1;
    int  iMonthLen = 0;

    // if no parameter provided, use current time:
    if (lMilliseconds == d_time_nan)
        lMilliseconds = std.date.UTCtoLocalTime(std.date.getUTCtime());
        
    // convert the milliseconds to number of sols since MARS_START:
    rSols = ((lMilliseconds - MARS_START) / MILLISECONDS_PER_SOL);
    rTime = ((lMilliseconds - MARS_START) - (rSols * MILLISECONDS_PER_SOL));
    rSols++;

    rTime    *= 0.9732442977008320022189969987579; // Adjust Earth time to Mars time
    iSpells   = cast(int)(rTime / MILLISECONDS_PER_HOUR);
    rTime    -= (iSpells * MILLISECONDS_PER_HOUR);
    iMoments  = cast(int)(rTime / MILLISECONDS_PER_MINUTE); 
    rTime    -= iMoments * MILLISECONDS_PER_MINUTE;
    iJiffies  = cast(int)(rTime / MILLISECONDS_PER_SECOND); 
    rTime    -= iJiffies * MILLISECONDS_PER_SECOND;
    ims       = cast(int)rTime;

    // which mir:
    if (rSols < 0.0L) // it's a negative mir
    {
        while (rSols < 0.0L)
        {
            iMir--;
            iMirLen  = getSolsInMir(iMir);
            rSols   += iMirLen;
        }
    }
    else // it's 0 or +ve mir:
    {
        iMirLen = getSolsInMir(iMir);

        // if we have more sols left than there are in current mir,
        // subtract the number of sols in the current mir from the remainder,
        // and go to the next mir:
        while (rSols > iMirLen)
        {
            rSols -= iMirLen;
            iMir++;
            iMirLen = getSolsInMir(iMir);
        }
    }

    // which month:
    iMonthLen = getSolsInMonth(iMonth, iMir);

    while (rSols > iMonthLen)
    {
        // goto next month:
        rSols -= iMonthLen;
        iMonth++;
        iMonthLen = getSolsInMonth(iMonth, iMir);
    }

    if (rSols == 0)
        rSols++;

    mdtDate.iMir     = iMir;
    mdtDate.iMonth   = iMonth;
    mdtDate.iSol     = cast(int)rSols;
    mdtDate.iSpells  = (iSpells  < 0 ? 0 : iSpells);
    mdtDate.iMoments = (iMoments < 0 ? 0 : iMoments);
    mdtDate.iJiffies = (iJiffies < 0 ? 0 : iJiffies);
    mdtDate.ims      = (ims < 0      ? 0 : ims);

    /+
    writefln("mir=%d, month=%d (\"%s\"), sol=%d, weekday=\"%s\"", 
             mdtDate.iMir, mdtDate.iMonth, getMarsMonthDarian(mdtDate.iMonth), 
             mdtDate.iSol, getSolNameDarian(mdtDate.iMonth));

    writefln("Spells=%d, Moments=%d, Jiffies=%d, ms=%d",
             mdtDate.iSpells, mdtDate.iMoments, mdtDate.iJiffies, mdtDate.ims);

    writefln("Earth lMilliseconds=%X (%d)", lMilliseconds, lMilliseconds);
    +/

    return mdtDate;
}

// Create an Earth datetime based on a Maritan datetime
d_time createMarsDate
(    
     in int iMir, 
     in int iMonth, 
     in int iSol,
     in int iSpells  = 0,
     in int iMoments = 0,
     in int iJiffies = 0,
     in int ims = 0
)
{
    long lMirMS = 0L;
    int  iSols  = 0;
    long lTime  = 0L;
    
    if (iMir > 0)
        for ( int i = 1; i <= iMir; i++ )
            iSols += getSolsInMir(i);
    
    if (iMonth > 0 && iMonth <= 24)    
        for ( int i = 1; i < iMonth; i++ )
            iSols += getSolsInMonth(i, iMir);
        
    if (iSol > 0 && iSol <= getSolsInMonth(iMonth, iMir))
        iSols += iSol;

    lMirMS = MARS_START;
    
    if (isLeapMir(iMir))
        lMirMS += (iSols - 1) * MILLISECONDS_PER_SOL; // (1000 * 60 * 60 * 24)
    else
        lMirMS += iSols * MILLISECONDS_PER_SOL;       // (1000 * 60 * 60 * 24)
        
    lTime  += iSpells   * MILLISECONDS_PER_HOUR;      // (1000 * 60 * 60)
    lTime  += iMoments  * MILLISECONDS_PER_MINUTE;    // (1000 * 60)
    lTime  += iJiffies  * MILLISECONDS_PER_SECOND;    // 1000
    lTime  += ims + 1;    // extra milliseconds added
    lTime  *= 1.02749125; // Maratian Time is a bit longer
    lMirMS += lTime;

    return lMirMS;
}

// Uses only the Darian Calendar
char[] toMarsDateString(in d_time dtDate)
{
    marsdatetime mdtDate = getMarsDate(dtDate);
    StringBuffer sBuffy;

    with (mdtDate)
    {
        sBuffy.setsize(28);
        sBuffy.allowoverflow(false);

        sBuffy.append(getAbbrevSolNameDarian(iSol), ", "c,
                      zfill(std.string.toString(iSol), 2), " "c,
                      getAbbrevMarsMonth(iMonth), " "c,
                      zfill(std.string.toString(iMir), 3), " "c,
                      zfill(std.string.toString(iSpells), 2), ":"c,
                      zfill(std.string.toString(iMoments), 2), ":"c,
                      zfill(std.string.toString(iJiffies), 2), "."c,
                      zfill(std.string.toString(ims), 3));
    }

    return sBuffy.getBuffer(); 
}

// For the Darian Calendar
// returns the month name given the month number (1..24):
char[] getMarsMonthDarian(in int iMonth)
{
    switch (iMonth)
    {
      case 1:  return "Sagittarius"c;
      case 2:  return "Dhanus"c;
      case 3:  return "Capricornus"c;
      case 4:  return "Makara"c;
      case 5:  return "Aquarius"c;
      case 6:  return "Khumba"c;
      case 7:  return "Pisces"c;
      case 8:  return "Mina"c;
      case 9:  return "Aries"c;
      case 10: return "Mesha"c;
      case 11: return "Taurus"c;
      case 12: return "Rishabha"c;
      case 13: return "Gemini"c;
      case 14: return "Mithuna"c;
      case 15: return "Cancer"c;
      case 16: return "Karka"c;
      case 17: return "Leo"c;
      case 18: return "Simha"c;
      case 19: return "Virgo"c;
      case 20: return "Kanya"c;
      case 21: return "Libra"c;
      case 22: return "Tula"c;
      case 23: return "Scorpius"c;
      case 24: return "Vrishika"c;
      default: return ""c;
    }
}

// For the Darian Calendar
// returns the month name given the month number (1..24):
char[] getAbbrevMarsMonth(in int iMonth)
{
    switch (iMonth)
    {
      case 1:  return "Sag"c;
      case 2:  return "Dha"c;
      case 3:  return "Cap"c;
      case 4:  return "Mak"c;
      case 5:  return "Aqr"c;
      case 6:  return "Kum"c;
      case 7:  return "Psc"c;
      case 8:  return "Min"c;
      case 9:  return "Ari"c;
      case 10: return "Mes"c;
      case 11: return "Tau"c;
      case 12: return "Ris"c;
      case 13: return "Gem"c;
      case 14: return "Mit"c;
      case 15: return "Can"c;
      case 16: return "Kar"c;
      case 17: return "Leo"c;
      case 18: return "Sim"c;
      case 19: return "Vir"c;
      case 20: return "Kan"c;
      case 21: return "Lib"c;
      case 22: return "Tul"c;
      case 23: return "Sco"c;
      case 24: return "Vri"c;
      default: return ""c;
    }
}

// For the Darian Calendar
// returns the sol name given the sol number (1..28):
char[] getSolNameDarian(in int iSolOfMonth)
{
    int iSolOfWeek = (iSolOfMonth - 1) % 7 + 1;

    switch (iSolOfWeek)
    {
      case 1: return "Sol Solis"c;
      case 2: return "Sol Lunae"c;
      case 3: return "Sol Martis"c;
      case 4: return "Sol Mercurii"c;
      case 5: return "Sol Jovis"c;
      case 6: return "Sol Veneris"c;
      case 7: return "Sol Saturni"c;
      default: return ""c;
    }
}

// For the Darian Calendar
char[] getAbbrevSolNameDarian(in int iSolOfMonth)
{
    int iSolOfWeek = (iSolOfMonth - 1) % 7 + 1;

    switch (iSolOfWeek)
    {
      case 1: return "Sol"c;
      case 2: return "Lun"c;
      case 3: return "Mar"c;
      case 4: return "Mer"c;
      case 5: return "Jov"c;
      case 6: return "Ven"c;
      case 7: return "Sat"c;
      default: return ""c;
    }
}

// Others not used here, but lefted in.
// returns the month name given the month number (1..24):
char[] getMarsMonthDarianDefrost(in int iMonth)
{
    switch (iMonth)
    {
      case 1:  return "Adir"c;
      case 2:  return "Bora"c;
      case 3:  return "Coan"c;
      case 4:  return "Deti"c;
      case 5:  return "Edal"c;
      case 6:  return "Flo"c;
      case 7:  return "Geor"c;
      case 8:  return "Heliba"c;
      case 9:  return "Idanon"c;
      case 10: return "Jowani"c;
      case 11: return "Kireal"c;
      case 12: return "Larno"c;
      case 13: return "Medior"c;
      case 14: return "Neturima"c;
      case 15: return "Ozulikan"c;
      case 16: return "Pasurabi"c;
      case 17: return "Rudiakel"c;
      case 18: return "Safundo"c;
      case 19: return "Tiunor"c;
      case 20: return "Ulasja"c;
      case 21: return "Vadeun"c;
      case 22: return "Wakumi"c;
      case 23: return "Xetual"c;
      case 24: return "Zungo"c;
      default: return ""c;
    }
}

// returns the month name given the month number (1..24):
char[] getMarsMonthDarianHensel(in int iMonth)
{
    switch (iMonth)
    {
      case 1:  return "Vernalis"c;
      case 2:  return "Duvernalis"c;
      case 3:  return "Trivernalis"c;
      case 4:  return "Quadrvernalis"c;
      case 5:  return "Pentvernalis"c;
      case 6:  return "Hexvernalis"c;
      case 7:  return "Aestas"c;
      case 8:  return "Duestas"c;
      case 9:  return "Triestas"c;
      case 10: return "Quadrestas"c;
      case 11: return "Pentestas"c;
      case 12: return "Hexestas"c;
      case 13: return "Autumnus"c;
      case 14: return "Duautumn"c;
      case 15: return "Triautumn"c;
      case 16: return "Quadrautumn"c;
      case 17: return "Pentautumn"c;
      case 18: return "Hexautumn"c;
      case 19: return "Unember"c;
      case 20: return "Duember"c;
      case 21: return "Triember"c;
      case 22: return "Quadrember"c;
      case 23: return "Pentember"c;
      case 24: return "Hexember"c;
      default: return ""c;
    }
}

// returns the month name given the month number (1..24):
char[] getMarsMonthUtopian(in int iMonth)
{
    switch (iMonth)
    { 
      case 1:  return "Sagittarius"c;
      case 2:  return "Aquila"c;
      case 3:  return "Capricornus"c;
      case 4:  return "Grus"c;
      case 5:  return "Aquarius"c;
      case 6:  return "Pegasus"c;
      case 7:  return "Pisces"c;
      case 8:  return "Cetus"c;
      case 9:  return "Aries"c;
      case 10: return "Perseus"c;
      case 11: return "Taurus"c;
      case 12: return "Orion"c;
      case 13: return "Gemini"c;
      case 14: return "Monoceros"c;
      case 15: return "Cancer"c;
      case 16: return "Hydra"c;
      case 17: return "Leo"c;
      case 18: return "Corvus"c;
      case 19: return "Virgo"c;
      case 20: return "Bootes"c;
      case 21: return "Libra"c;
      case 22: return "Lupus"c;
      case 23: return "Scorpius"c;
      case 24: return "Ophiuchus"c;
      default: return ""c;
    }
}

// returns the sol name given the sol number (1..28):
char[] getSolNameDarianDefrost(in int iSolOfMonth)
{
    int iSolOfWeek = (iSolOfMonth - 1) % 7 + 1;

    switch (iSolOfWeek)
    {
      case 1: return "Axatisol"c;
      case 2: return "Benasol"c;
      case 3: return "Ciposol"c;
      case 4: return "Domesol"c;
      case 5: return "Erjasol"c;
      case 6: return "Fulisol"c;
      case 7: return "Gavisol"c;
      default: return ""c;
    }
}

// returns the sol name given the sol number (1..28):
char[] getSolNameUtopian(in int iSolOfMonth)
{
    int iSolOfWeek = (iSolOfMonth - 1) % 7 + 1;

    switch (iSolOfWeek)
    {
      case 1: return "Sunsol"c;
      case 2: return "Phobosol"c;
      case 3: return "Deimosol"c;
      case 4: return "Terrasol"c;
      case 5: return "Venusol"c;
      case 6: return "Mercurisol"c;
      case 7: return "Jovisol"c;
      default: return ""c;
    }
}

unittest
{
    // d_time remains a valid Earth datetime that can 
    // on the fly be converted to a Mars datetime.
    d_time       dtDate = d_time_nan;
    marsdatetime mdtDate;

    writefln("M2E = a Mars datetime coverted to an Earth datetime."c);
    writefln("E2M = a Earth datetime coverted to an Mars datetime."c);
    writefln();

    // Mars ="Mar, 10 Sim 210 20:25:17.000" 
    // Earth="Thu, 14 Jul 2005 19:29:47 UTC"
    dtDate = createMarsDate(210, 18, 10, 20, 25, 17, 0);
    writefln("M2E Mars=\"%s\" Earth=\"%s\""c, toMarsDateString(dtDate), toUTCString(dtDate));
    assert(toMarsDateString(dtDate) == "Mar, 10 Sim 210 20:25:17.000"c);
    writefln();

    // UTC(1975, 11, 17, 23, 46, 12, 0)
    dtDate = parse("11/17/1975 23:46:12 UTC"c);

    // Viking 1 Lander: Landed on Mars at Chryse Planitia
    // ---------------------------------------------------
    // Mars ="Sat, 14 Min 195 18:40:26.951" 
    // Earth="Tue, 20 Jul 1976 11:53:06 UTC"
    writefln("Viking 1 Lander: Landed on Mars at Chryse Planitia"c);
    dtDate = parse("07/20/1976 11:53:06 UTC"c);
    writefln("E2M Mars=\"%s\" Earth=\"%s\""c, toMarsDateString(dtDate), toUTCString(dtDate));
    assert(toMarsDateString(dtDate) == "Sat, 14 Min 195 18:40:26.951"c);
    writefln();

    // start date for Martian northern vernal equinox in 1609, the year
    // Astronomy Novia was published by Johannes Kepler, and also the year
    // the telescope was first used for astronomy, by Galileo Galilei:
    // 1609 Mar 11 08:40:48.400
    // (jscript) const MARS_START = Date.UTC(1609, 02, 11, 18, 40, 48, 400).valueOf();
    writefln("Telescope was first used for astronomy, by Galileo Galilei"c);
    dtDate = createMarsDate(0, 1, 1, 0, 0, 0, 0);
    writefln("M2E Mars=\"%s\" Earth=\"%s\""c, toMarsDateString(dtDate), toUTCString(dtDate));
    assert(toUTCString(dtDate) == "Wed, 11 Mar 1609 18:40:48 UTC"c);

    dtDate = std.date.parse("03/11/1609 18:40:48 UTC"c) + (400 * TicksPerMs);
    writefln("E2M Mars=\"%s\" Earth=\"%s\""c, toMarsDateString(dtDate), toUTCString(dtDate));
    assert(toMarsDateString(dtDate) == "Sol, 01 Sag 000 00:00:00.000"c);
    writefln();

    // Spirit, 2003 Mars Exploration Rover 2 lands on Mars at 
    // 14.57 degrees south latitude and 175.47 degrees east longitude.
    // ---------------------------------------------------------------
    // Mars ="Jov, 26 Tul 209 03:34:41.686"
    // Earth="Sun, 04 Jan 2004 04:35:03 UTC"
    writefln("Spirit, 2003 Mars Exploration Rover 2 lands on Mars"c);
    dtDate = createMarsDate(209, 22, 26, 3, 34, 41, 686);
    writefln("M2E Mars=\"%s\" Earth=\"%s\""c, toMarsDateString(dtDate), toUTCString(dtDate));
    assert(toUTCString(dtDate) == "Sun, 04 Jan 2004 04:35:03 UTC"c);

    dtDate = std.date.parse("01/04/2004 04:35:03 UTC"c) + (400 * TicksPerMs);
    writefln("E2M Mars=\"%s\" Earth=\"%s\""c, toMarsDateString(dtDate), toUTCString(dtDate));
    assert(toMarsDateString(dtDate) == "Jov, 26 Tul 209 03:34:41.686"c);
    writefln();

    // Opportunity, 2003 Mars Exploration Rover 1 lands on Mars at
    // 1.95 degrees south latitude and 354.47 degrees east longitude.
    // ---------------------------------------------------------------
    // Mars ="Mer, 18 Sco 209 14:34:46.033"
    // Earth="Sun, 25 Jan 2004 05:05:01 UTC"
    writefln("Opportunity, 2003 Mars Exploration Rover 1 lands on Mars"c);
    dtDate = createMarsDate(209, 23, 18, 14, 34, 46, 33);
    writefln("M2E Mars=\"%s\" Earth=\"%s\""c, toMarsDateString(dtDate), toUTCString(dtDate));
    assert(toUTCString(dtDate) == "Sun, 25 Jan 2004 05:05:01 UTC"c);

    dtDate = std.date.parse("01/25/2004 05:05:01 UTC"c) + (400 * TicksPerMs);
    writefln("E2M Mars=\"%s\" Earth=\"%s\""c, toMarsDateString(dtDate), toUTCString(dtDate));
    assert(toMarsDateString(dtDate) == "Mer, 18 Sco 209 14:34:46.033"c);
    writefln();

    // MKoD's Dyriia birthdate
    // ------------------------
    // Mars ="Sol, 01 Ari 229 02:10:28.854" 
    // Earth="Sun, 15 Jul 2040 02:30:55 UTC"
    writefln("MKoD's Dyriia birthdate"c);
    dtDate = createMarsDate(229, 9, 1, 2, 10, 28, 854);
    writefln("M2E Mars=\"%s\" Earth=\"%s\""c, toMarsDateString(dtDate), toUTCString(dtDate));
    assert(toUTCString(dtDate) == "Sun, 15 Jul 2040 02:30:55 UTC"c);

    dtDate = std.date.parse("07/15/2040 02:30:55 UTC"c);
    writefln("E2M Mars=\"%s\" Earth=\"%s\""c, toMarsDateString(dtDate), toUTCString(dtDate));
    assert(toMarsDateString(dtDate) == "Sol, 01 Ari 229 02:10:28.854"c);
    writefln();

    // Dyriia currently twenty-five in Earth years / thirteen in Martian mirs.
    writefln("Dyriia currently twenty-five in Earth years / thirteen in Martian mirs"c);
    dtDate = std.date.parse("07/15/2065 02:30:55 UTC"c);
    writefln("E2M Mars=\"%s\" Earth=\"%s\""c, toMarsDateString(dtDate), toUTCString(dtDate));
    assert(toMarsDateString(dtDate) == "Ven, 27 Can 242 18:49:23.006"c);
    dtDate = createMarsDate(242, 15, 27, 18, 49, 23, 6);
    writefln("M2E Mars=\"%s\" Earth=\"%s\""c, toMarsDateString(dtDate), toUTCString(dtDate));
    assert(toUTCString(dtDate) == "Wed, 15 Jul 2065 02:30:55 UTC"c);
}

debug( marsdate )
{
int main()
{
	writefln();
    writefln("Compiled and tested with: \"%s v%d.%03d\"", scomp.name, scomp.version_major, scomp.version_minor);
    writefln("marsdate.unittest finished!"c); 

    return 0;
}
}